from __future__ import absolute_import
from collections import OrderedDict

# Somehow find a way to enforce these keys to be bytestrings?

ENCODERS = OrderedDict()

try:
  import msgpack

  ENCODERS[b'msgpack'] = msgpack
except ImportError:
  pass

try:
  import json

  ENCODERS[b'json'] = json
except ImportError:
  pass

try:
  import erlpack

  class _erlpack:
    dumps = erlpack.pack
    loads = erlpack.unpack
  
  ENCODERS[b'erlpack'] = _erlpack
except ImportError:
  pass



from libc.stdint cimport uint32_t, uint8_t

cdef class Response:
    def __cinit__(self, uint8_t flags, uint32_t seq, bytes data):
        self.flags = flags
        self.seq = seq
        self.data = data

cdef class Request:
    def __cinit__(self, uint8_t flags, uint32_t seq, bytes data):
        self.flags = flags
        self.seq = seq
        self.data = data

cdef class Push:
    def __cinit__(self, uint8_t flags, bytes data):
        self.flags = flags
        self.data = data

cdef class Ping:
    def __cinit__(self, uint8_t flags, uint32_t seq):
        self.flags = flags
        self.seq = seq

cdef class Pong:
    def __cinit__(self, uint8_t flags, uint32_t seq):
        self.flags = flags
        self.seq = seq

cdef class Hello:
    def __cinit__(self, uint8_t flags, uint8_t version, list supported_encodings, list supported_compressions):
        self.flags = flags
        self.version = version
        self.supported_encodings = supported_encodings
        self.supported_compressions = supported_compressions

cdef class HelloAck:
    def __cinit__(self, uint8_t flags, uint32_t ping_interval, bytes selected_encoding, bytes selected_compression):
        self.flags = flags
        self.ping_interval = ping_interval
        self.selected_encoding = selected_encoding
        self.selected_compression = selected_compression

cdef class GoAway:
    def __cinit__(self, uint8_t flags, uint16_t code, bytes reason):
        self.flags = flags
        self.code = code
        self.reason = reason

cdef class Error:
    def __cinit__(self, uint8_t flags, uint16_t code, uint32_t seq, bytes data):
        self.flags = flags
        self.code = code
        self.seq = seq
        self.data = data



from libc.stdint cimport uint32_t, uint8_t, uint16_t

cdef class Response:
    cdef readonly uint8_t flags
    cdef readonly uint32_t seq
    cdef readonly bytes data

cdef class Request:
    cdef readonly uint8_t flags
    cdef readonly uint32_t seq
    cdef readonly bytes data

cdef class Push:
    cdef readonly uint8_t flags
    cdef readonly bytes data

cdef class Ping:
    cdef readonly uint8_t flags
    cdef readonly uint32_t seq

cdef class Pong:
    cdef readonly uint8_t flags
    cdef readonly uint32_t seq

cdef class Hello:
    cdef readonly uint8_t flags
    cdef readonly uint8_t version
    cdef readonly list supported_encodings
    cdef readonly list supported_compressions

cdef class HelloAck:
    cdef readonly uint8_t flags
    cdef readonly uint32_t ping_interval
    cdef readonly bytes selected_encoding
    cdef readonly bytes selected_compression

cdef class GoAway:
    cdef readonly uint8_t flags
    cdef readonly uint16_t code
    cdef readonly bytes reason

cdef class Error:
    cdef readonly uint8_t flags
    cdef readonly uint16_t code
    cdef readonly uint32_t seq
    cdef readonly bytes data



import errno
import logging
import socket

import gevent
from gevent.event import AsyncResult, Event

from libc.stdint cimport uint32_t, uint8_t

from toku.exceptions import NoEncoderAvailable, ConnectionTerminated, TokuDecoderError, StreamDefunct, \
    NotClientException, NotServerException, TokuErrorReceived
from opcodes cimport Request, Response, Ping, Pong, Push, Hello, GoAway, HelloAck, Error

from socket_watcher cimport SocketWatcher
from stream_handler cimport TokuStreamHandler

cdef size_t OUTBUF_MAX = 65535


class CloseReasons(object):
    NORMAL = 0
    PING_TIMEOUT = 1
    UNKNOWN_ENCODER = 2
    NO_MUTUAL_ENCODERS = 3
    DIDNT_STOP_IN_TIME = 4
    DECODER_ERROR = 5

cdef class TokuSocketSession:
    def __cinit__(self, object sock, object encoders, bint is_client=True, object on_request=None, object on_push=None):
        self._is_client = is_client
        self._stream_handler = TokuStreamHandler()
        self._sock = sock
        self._watcher = SocketWatcher(self._sock.fileno())
        self._inflight_requests = {}
        self._available_encoders = encoders
        self._available_compressors = {}
        self._shutdown_event = Event()
        self._close_event = Event()
        self._ready_event = Event()
        self._ping_interval = 5
        self._is_ready = False
        self._shutting_down = False

        if not self._is_client:
            self._on_request = on_request

        self._on_push = on_push

        gevent.spawn(self._ping_loop)
        gevent.spawn(self._run_loop)

        if is_client:
            self._send_hello()

    cpdef set_push_handler(self, object push_handler):
        self._on_push = None

    cdef _resume_sending(self):
        if self._sock is None:
            return

        if self._stream_handler.write_buffer_len() == 0:
            return

        self._watcher.switch_if_write_unblocked()

    cdef shutdown(self):
        if self._shutting_down:
            return

        self._shutting_down = True

    cdef void _cleanup_socket(self):
        sock = self._sock
        self._sock = None
        if sock:
            sock.close()

        # Request a switch so that the socket event loop will exit.
        if sock:
            self._watcher.request_switch()

        self._cleanup_inflight_requests(ConnectionTerminated())

        # Unblock anything waiting on ready event.
        if not self._ready_event.is_set():
            self._ready_event.set()

        # Unblock anything waiting on stop event.
        if not self._shutdown_event.is_set():
            self._shutdown_event.set()

        # Unblock anything waiting on the close event.
        if not self._close_event.is_set():
            self._close_event.set()

    cpdef close(self, uint8_t code=CloseReasons.NORMAL, bytes reason=None,
                block=False, block_timeout=None, close_timeout=None, bint via_remote_goaway=False):

        if not self._shutting_down:
            self._shutting_down = True

            # Unblock anything waiting on ready event.
            if not self._ready_event.is_set():
                self._ready_event.set()

            # Unblock anything waiting on stop event.
            if not self._shutdown_event.is_set():
                self._shutdown_event.set()

            # Send goaway to the remote node.
            if not via_remote_goaway and self._sock:
                self._stream_handler.send_goaway(0, code, reason)

            # Spawn a greenlet that will wait
            gevent.spawn(self._close_timeout)

        # If we are blocking, wait on close event to succeed.
        if block:
            self.join(block_timeout)

    cpdef _close_timeout(self):
        if self._close_event.wait(self._ping_interval):
            return

        self._cleanup_socket()

    cpdef join(self, timeout=None):
        self._close_event.wait(timeout=timeout)

    cpdef terminate(self):
        self._cleanup_socket()

    cdef _cleanup_inflight_requests(self, close_exception):
        requests = self._inflight_requests.values()
        self._inflight_requests.clear()

        for request in requests:
            if isinstance(request, AsyncResult):
                request.set_exception(close_exception)

    cdef bint await_ready(self) except 1:
        if not self._is_ready:
            self._ready_event.wait()

    cdef bint is_ready(self):
        return self._is_ready

    cdef _encode_data(self, object data):
        if not self._is_ready:
            self._ready_event.wait()

        if not self._encoder_dumps:
            raise NoEncoderAvailable()

        return 0, bytes(self._encoder_dumps(data), encoding="utf-8")

    cdef _decode_data(self, uint8_t flags, object data):
        if not self._is_ready:
            self._ready_event.wait()

        if not self._encoder_loads:
            raise NoEncoderAvailable()

        return bytes(self._encoder_loads(data), encoding="utf-8")

    cpdef object send_request(self, object data):
        if not self._is_client:
            raise NotClientException()

        if self.defunct():
            raise StreamDefunct

        cdef bytes encoded_data
        cdef uint8_t flags
        flags, encoded_data = self._encode_data(data)

        result = AsyncResult()
        cdef uint32_t seq = self._stream_handler.send_request(flags, encoded_data)
        self._inflight_requests[seq] = result
        self._resume_sending()
        return result

    cpdef object send_push(self, object data):
        if self.defunct():
            raise StreamDefunct()

        cdef bytes encoded_data
        cdef uint8_t flags
        flags, encoded_data = self._encode_data(data)

        self._stream_handler.send_push(flags, encoded_data)
        self._resume_sending()

    cpdef object send_response(self, uint32_t seq, object data):
        if self._is_client:
            raise NotServerException()

        if self.defunct():
            return

        cdef bytes encoded_data
        cdef uint8_t flags

        request = self._inflight_requests.pop(seq, None)
        if request:
            flags, encoded_data = self._encode_data(data)
            self._stream_handler.send_response(flags, seq, encoded_data)
            self._resume_sending()

        else:
            logging.error('Sending response for unknown seq %s', seq)

        return None

    cpdef object send_ping(self):
        if self.defunct():
            raise StreamDefunct()

        result = AsyncResult()
        cdef uint32_t seq = self._stream_handler.send_ping(0)
        self._inflight_requests[seq] = result
        self._resume_sending()
        return result

    cpdef object _send_hello_ack(self, bytes selected_encoding, bytes selected_compressor):
        if self._is_client:
            raise NotServerException()

        if self.defunct():
            raise StreamDefunct()

        self._stream_handler.send_hello_ack(0, int(self._ping_interval * 1000), selected_encoding, selected_compressor)

    cpdef object _send_hello(self):
        if not self._is_client:
            raise NotClientException()

        self._stream_handler.send_hello(0, list(self._available_encoders.keys()), list(self._available_compressors.keys()))

    cdef _handle_ping_timeout(self):
        self.close(code=CloseReasons.PING_TIMEOUT)

    cdef _handle_data_received(self, data):
        try:
            events = self._stream_handler.on_bytes_received(data)

        except TokuDecoderError as e:
            print e
            return self.close(code=CloseReasons.DECODER_ERROR)

        if not events:
            return

        if self._is_client:
            self._handle_client_events(events)

        else:
            self._handle_server_events(events)

        self._resume_sending()

    cdef _handle_client_events(self, list events):
        for event in events:
            if isinstance(event, Response):
                self._handle_response(event)

            elif isinstance(event, Push):
                self._handle_push(event)

            elif isinstance(event, Ping):
                self._handle_ping(event)

            elif isinstance(event, Pong):
                self._handle_pong(event)

            elif isinstance(event, GoAway):
                self._handle_go_away(event)

            elif isinstance(event, HelloAck):
                self._handle_hello_ack(event)

            elif isinstance(event, Error):
                self._handle_error(event)

    cdef _handle_server_events(self, list events):
        for event in events:
            if isinstance(event, Request):
                self._handle_request(event)

            elif isinstance(event, Push):
                self._handle_push(event)

            elif isinstance(event, Ping):
                self._handle_ping(event)

            elif isinstance(event, Pong):
                self._handle_pong(event)

            elif isinstance(event, GoAway):
                self._handle_go_away(event)

            elif isinstance(event, Hello):
                self._handle_hello(event)

    cdef _handle_request(self, Request request):
        if self._on_request:
            # In this case, we set the inflight requests to the given request. That way send_response
            # will know if the seq is valid or not.
            request.data = self._decode_data(request.flags, request.data)
            self._inflight_requests[request.seq] = request
            response = self._on_request(request, self)
            # If a response is given, we can return it to the sender right away.
            # Otherwise, it's the responsibility of the `_on_request` handler to eventually
            # call `send_response`.
            if response is not None:
                self.send_response(request.seq, response)

    cdef _handle_response(self, Response response):
        request = self._inflight_requests.pop(response.seq, None)
        if request:
            try:
                # If we've gotten a response for a request we've made.
                request.set(self._decode_data(response.flags, response.data))
            except Exception as e:
                request.set_exception(e)

    cdef _handle_push(self, Push push):
        if self._on_push:
            push.data = self._decode_data(push.flags, push.data)
            self._on_push(push, self)

    cdef _handle_ping(self, Ping ping):
        # Nothing to do here - the stream handler handles sending pongs back for us.
        pass

    cdef _handle_pong(self, Pong pong):
        ping_request = self._inflight_requests.pop(pong.seq, None)
        if ping_request:
            ping_request.set(pong)

    cdef _handle_hello(self, Hello hello):
        encoding, encoder = self._pick_best_encoding(hello.supported_encodings)

        if not encoding:
            self.close(reason=CloseReasons.NO_MUTUAL_ENCODERS)

        else:
            self._encoder_dumps = encoder.dumps
            self._encoder_loads = encoder.loads
            self._encoding = encoding
            self._send_hello_ack(encoding, None)
            self._is_ready = True
            self._ready_event.set()

    cdef _handle_hello_ack(self, HelloAck hello_ack):
        encoder = self._available_encoders.get(hello_ack.selected_encoding)
        if not encoder:
            self.close(code=CloseReasons.UNKNOWN_ENCODER, reason=b'Unknown encoding %s' % hello_ack.selected_encoding)

        else:
            self._ping_interval = int(hello_ack.ping_interval / 1000)
            self._encoder_dumps = encoder.dumps
            self._encoder_loads = encoder.loads
            self._encoding = hello_ack.selected_encoding
            self._is_ready = True
            self._ready_event.set()

    cdef _handle_go_away(self, GoAway go_away):
        print "Got goaway on seq", self._stream_handler.current_seq(), go_away.code
        self.close(via_remote_goaway=True, reason=go_away.reason, code=go_away.code)

    cdef _handle_error(self, Error error):
        request = self._inflight_requests.pop(error.seq, None)
        if request:
            request.set_exception(TokuErrorReceived(error.code, error.data))

    cdef _pick_best_encoding(self, list encodings):
        for encoding in encodings:
            encoder = self._available_encoders.get(encoding)
            if encoder:
                return encoding, encoder

        return None, None

    cpdef _ping_loop(self):
        while self._sock:
            ping_result = self.send_ping()
            if self._shutdown_event.wait(self._ping_interval):
                return

            if not ping_result.ready():
                self._handle_ping_timeout()
                return

    cpdef _run_loop(self):
        loop = gevent.get_hub().loop
        io = loop.io
        cdef int MAXPRI = loop.MAXPRI
        cdef int READ = 1
        cdef int WRITE = 2
        cdef bint write_watcher_started = False
        cdef bint sock_should_write = False
        cdef bint did_empty_buffer = False
        cdef object sock_recv = self._sock.recv
        cdef object sock_send = self._sock.send
        cdef object watcher_mark_ready = (<object> self._watcher).mark_ready
        cdef size_t write_bytes_remaining

        sock_read_watcher = io(self._watcher.sock_fileno, READ)
        sock_write_watcher = io(self._watcher.sock_fileno, WRITE)

        try:
            sock_read_watcher.start(watcher_mark_ready, self._watcher.sock_fileno, True)

            while self._sock:
                if write_watcher_started == False and self._stream_handler.write_buffer_len() > 0:
                    sock_write_watcher.start(watcher_mark_ready, self._watcher.sock_fileno, False)

                self._watcher.wait()
                if not self._sock:
                    return

                if self._watcher.sock_read_ready:
                    try:
                        data = sock_recv(65536)
                    except socket.error as e:
                        if e.errno in (errno.EAGAIN, errno.EINPROGRESS):
                            continue

                        data = None

                    if not data:
                        self.close()
                        self._cleanup_socket()
                        return

                    else:
                        self._handle_data_received(data)

                # We should attempt to write, if the watcher has notified us that the socket is ready
                # to accept more data, or we aren't write blocked yet - and we have a write buffer.
                sock_should_write = self._watcher.sock_write_ready or (
                    not self._watcher.sock_write_blocked and self._stream_handler.write_buffer_len() > 0
                )

                if sock_should_write:
                    try:
                        bytes_written = sock_send(self._stream_handler.write_buffer_get_bytes(OUTBUF_MAX, False))
                    except socket.error as e:
                        if e.errno in (errno.EAGAIN, errno.EINPROGRESS):
                            bytes_written = 0

                        else:
                            self.close()
                            self._cleanup_socket()

                    # No bytes have been written. It's safe to assume the socket is still (somehow) blocked
                    # and we don't need to do anything.
                    if not bytes_written:
                        self._watcher.sock_write_blocked = True
                        continue

                    write_bytes_remaining = self._stream_handler.write_buffer_consume_bytes(bytes_written)
                    # Did we completely write the buffer? If so - the socket isn't blocked anymore.
                    self._watcher.sock_write_blocked = write_bytes_remaining > 0

                    # If no data is available to send - we can stop the watcher. Otherwise,
                    # we will leave the watcher open, so the data filled into the buffer
                    # will attempt to be written upon the next tick of the event loop.
                    if write_bytes_remaining == 0:
                        sock_write_watcher.stop()
                        write_watcher_started = False

                # If we are shutting down, have an empty write buffer, and no more in-flight requests,
                # it's safe to terminate the socket.
                if self._shutting_down and self._stream_handler.write_buffer_len() == 0 and not self._inflight_requests:
                    print 'cleaning up socket because we are drained'
                    self._cleanup_socket()

                self._watcher.reset()

        finally:
            sock_read_watcher.stop()
            sock_write_watcher.stop()

    cdef bint defunct(self):
        if self._sock is None:
            return True

        if self._shutting_down:
            return True

        return False



import random

cdef class Backoff:
  def __init__(self, min_delay=0.5, max_delay=None, jitter=True):
    self._min = min_delay
    if max_delay is None:
      max_delay = min_delay * 10
    
    self._max = max_delay
    self._jitter = jitter

    self._current = self._min
    self._fails = 0
  
  cdef int fails(self):
    return self._fails
  
  cdef float current(self):
    return self._current

  cdef void succeed(self):
    self._fails = 0
    self._current = self._min
  
  cdef float fail(self):
    self._fails += 1
    delay = self._current * 2
    if self._jitter:
      delay *= random.random()
    
    self._current += delay

    if self._max:
      self._current = min(self._current, self._max)

    self._current = round(self._current, 2)
    return self._current



from cpython cimport *
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy
from libc.stdint cimport uint32_t, uint8_t, uint16_t

from exceptions import TokuDecoderError
cimport toku_c
cimport opcodes

cdef size_t BBS = 1024 * 1024 * 2 # BIG_BUF_SIZE
cdef size_t IBS = 1024 * 512 # INITIAL_BUFFER_SIZE
# TODO: Move to header
cdef uint32_t SEQ_MAX = (2 ** 32) - 2

cdef inline void _reset_buffer(toku_c.toku_buffer_t *toku_buffer):
  """
  Ensures the buffer is in a reset state.
  """
  if toku_buffer.buf != NULL:
    toku_buffer.length = 0
  
  else:
    toku_buffer.buf = <char*> malloc(IBS)
    if toku_buffer.buf == NULL:
      raise MemoryError('Unable to allocate buffer')
    
    toku_buffer.allocated_size = IBS
    toku_buffer.length = 0
  

cdef inline void _free_big_buffer(toku_c.toku_buffer_t *toku_buffer):
  """
  Frees the buffer if it's grown over `BIG_BUF_SIZE`
  """
  if toku_buffer.allocated_size >= BBS:
    free(toku_buffer.buf)
    toku_buffer.buf = NULL
    toku_buffer.length = 0
    toku_buffer.allocated_size = 0
  
  else:
    toku_buffer.length = 0


cdef inline bytes _get_payload_from_decode_buffer(toku_c.toku_decode_buffer_t *decode_buffer):
  """
  Get the payload part of the decode buffer. Copies the data after the headers in the buffer into a PyBytes object.
  """
  cdef char *buf = decode_buffer.toku_buffer.buf + decode_buffer.header_size
  cdef size_t size = toku_c.toku_get_data_payload_size(decode_buffer)
  if size == 0:
    return None
  
  return PyBytes_FromStringAndSize(buf, size)

cdef class TokuStreamHandler:
  def __cinit__(self):
    self.seq = 0
    self.write_buffer_position = 0

    _reset_buffer(&self.write_buffer)
    _reset_buffer(&self.decode_buffer.toku_buffer)

  def __dealloc__(self):
    if self.write_buffer.buf != NULL:
      free(self.write_buffer.buf)
    
    if self.decode_buffer.toku_buffer.buf != NULL:
      free(self.decode_buffer.toku_buffer.buf)

  cpdef uint32_t current_seq(self):
    return self.seq

  cdef inline uint32_t next_seq(self):
    cdef uint32_t next_seq
    self.seq += 1
    next_seq = self.seq
    if next_seq >= SEQ_MAX:
      self.seq = 0
      next_seq = 0
    
    return next_seq

  cpdef uint32_t send_ping(self, uint8_t flags) except 0:
    cdef int rv
    cdef uint32_t seq = self.next_seq()
    rv = toku_c.toku_append_ping(&self.write_buffer, flags, seq)
    if rv < 0:
      raise MemoryError()
    
    return seq

  cpdef uint32_t send_pong(self, uint8_t flags, uint32_t seq) except 0:
    cdef int rv
    rv = toku_c.toku_append_pong(&self.write_buffer, flags, seq)
    if rv < 0:
      raise MemoryError()
    
    return 1

  cpdef uint32_t send_request(self, uint8_t flags, bytes data) except 0:
    cdef int rv
    cdef uint32_t seq = self.next_seq()
    cdef char *buffer
    cdef size_t size

    rv = PyBytes_AsStringAndSize(data, &buffer, <Py_ssize_t*> &size)
    if rv < 0:
      raise TypeError()
    
    rv = toku_c.toku_append_request(&self.write_buffer, flags, seq, size, <const char*> buffer)
    if rv < 0:
      raise MemoryError()
    
    return seq

  cpdef uint32_t send_push(self, uint8_t flags, bytes data) except 0:
    cdef int rv
    cdef char *buffer
    cdef size_t size

    rv = PyBytes_AsStringAndSize(data, &buffer, <Py_ssize_t*> &size)
    if rv < 0:
      raise TypeError()
    
    rv = toku_c.toku_append_push(&self.write_buffer, flags, size, <const char*> buffer)
    if rv < 0:
      raise MemoryError()

    return 1

  cpdef uint32_t send_hello(self, uint8_t flags, list supported_encodings, list supported_compressors) except 0:
    cdef int rv
    cdef char *buffer
    cdef size_t size

    for b in supported_encodings:
      print("encoding: ", b)
    for b in supported_compressors:
      print("compressor: ", b)

    cdef bytes data = bytes(b'%s|%s' % (
      b','.join([bytes(b) for b in supported_encodings]),
      b','.join([bytes(b) for b in supported_compressors])
    ))

    rv = PyBytes_AsStringAndSize(data, &buffer, <Py_ssize_t*> &size)
    if rv < 0:
      raise TypeError()
    
    rv = toku_c.toku_append_hello(&self.write_buffer, flags, size, <const char*> buffer)
    if rv < 0:
      raise MemoryError()
    
    return 1

  cpdef uint32_t send_hello_ack(self, uint8_t flags, uint32_t ping_interval, bytes selected_encoding, bytes selected_compressor) except 0:
    cdef int rv
    cdef char *buffer
    cdef size_t size
    cdef bytes data = bytes(b'%s|%s' % (
      selected_encoding,
      selected_compressor or b''
    ))

    rv = PyBytes_AsStringAndSize(data, &buffer, <Py_ssize_t*> &size)
    if rv < 0:
      raise MemoryError()
    
    rv = toku_c.toku_append_hello_ack(&self.write_buffer, flags, ping_interval, size, <const char*> buffer)
    if rv < 0:
      raise MemoryError()
    
    return 1

  cpdef uint32_t send_response(self, uint8_t flags, uint32_t seq, bytes data) except 0:
    cdef int rv
    cdef char *buffer
    cdef size_t size

    rv = PyBytes_AsStringAndSize(data, &buffer, <Py_ssize_t*> &size)
    if rv < 0:
      raise TypeError()
    
    rv = toku_c.toku_append_response(&self.write_buffer, flags, seq, size, <const char*> buffer)
    if rv < 0:
      raise MemoryError()
  
    return 1

  cpdef uint32_t send_error(self, uint8_t flags, uint16_t code, uint32_t seq, bytes data) except 0:
    cdef int rv
    cdef char *buffer = NULL
    cdef size_t size = 0

    if data:
      rv = PyBytes_AsStringAndSize(data, &buffer, <Py_ssize_t*> &size)
      if rv < 0:
        raise TypeError()
      
    rv = toku_c.toku_append_error(&self.write_buffer, flags, code, seq, size, <const char*> buffer)
    if rv < 0:
      raise MemoryError()
    
    return 1

  cpdef uint32_t send_goaway(self, uint8_t flags, uint16_t code, bytes reason) except 0:
    cdef int rv
    cdef char *buffer = NULL
    cdef size_t size = 0

    if reason:
      rv = PyBytes_AsStringAndSize(reason, &buffer, <Py_ssize_t*> &size)
      if rv < 0:
        raise TypeError()
      
    rv = toku_c.toku_append_goaway(&self.write_buffer, flags, code, size, <const char*> buffer)
    if rv < 0:
      raise MemoryError()
    
    return 1

  cpdef size_t write_buffer_len(self):
    return self.write_buffer.length - self.write_buffer_position

  cpdef bytes write_buffer_get_bytes(self, size_t length, bint consume=True):
    cdef size_t buffer_len_remaining = self.write_buffer_len()
    if length > buffer_len_remaining:
      length = buffer_len_remaining
    
    if length == 0:
      return None
    
    cdef bytes write_buffer = PyBytes_FromStringAndSize(self.write_buffer.buf + self.write_buffer_position, length)
    if consume:
      self.write_buffer_position += length
      self._reset_or_compact_write_buf()
    
    return write_buffer

  cpdef size_t write_buffer_consume_bytes(self, size_t length):
    cdef size_t buffer_len_remaining = self.write_buffer_len()
    if length > buffer_len_remaining:
      length = buffer_len_remaining
    
    self.write_buffer_position += length
    self._reset_or_compact_write_buf()
    return buffer_len_remaining - length

  cpdef list on_bytes_received(self, bytes data):
    cdef size_t size = PyBytes_Size(data)
    cdef char *buf = PyBytes_AS_STRING(data)
    cdef size_t consumed = 0
    cdef toku_c.toku_decoder_status decoder_status

    cdef list received_payloads = []
    while size > 0:
      decoder_status = toku_c.toku_decoder_read_data(&self.decode_buffer, size, buf, &consumed)
      if decoder_status < 0:
        self._reset_decode_buf()
        raise TokuDecoderError('The decoder failed with status %s' &decoder_status)
      
      if decoder_status == toku_c.TOKU_DECODE_NEEDS_MORE:
        break
      
      elif decoder_status == toku_c.TOKU_DECODE_COMPLETE:
        received_payloads.append(self._consume_decode_buffer())
      
      else:
        raise TokuDecoderError('Unhandled decoder status %s' % decoder_status)
      
      size -= consumed
      buf += consumed
      consumed = 0
    
    return received_payloads
  
  cdef object _consume_decode_buffer(self):
    cdef uint8_t opcode = self.decode_buffer.opcode
    cdef object response

    if opcode == toku_c.TOKU_OP_RESPONSE:
      response = opcodes.Response(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_seq(&self.decode_buffer),
        _get_payload_from_decode_buffer(&self.decode_buffer)
      )
    
    elif opcode == toku_c.TOKU_OP_REQUEST:
      response = opcodes.Request(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_seq(&self.decode_buffer),
        _get_payload_from_decode_buffer(&self.decode_buffer)
      )
    
    elif opcode == toku_c.TOKU_OP_PUSH:
      response = opcodes.Push(
        toku_c.toku_get_flags(&self.decode_buffer),
        _get_payload_from_decode_buffer(&self.decode_buffer)
      )
    
    elif opcode == toku_c.TOKU_OP_PING:
      response = opcodes.Ping(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_seq(&self.decode_buffer)
      )
    
    elif opcode == toku_c.TOKU_OP_PONG:
      response = opcodes.Pong(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_seq(&self.decode_buffer)
      )
    
    elif opcode == toku_c.TOKU_OP_GOAWAY:
      response = opcodes.GoAway(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_seq(&self.decode_buffer),
        _get_payload_from_decode_buffer(&self.decode_buffer)
      )
    
    elif opcode == toku_c.TOKU_OP_HELLO:
      payload = _get_payload_from_decode_buffer(&self.decode_buffer)
      supported_encodings, supported_compressions = payload.split(b'|')

      response = opcodes.Hello(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_flags(&self.decode_buffer),
        supported_encodings.split(b'|'),
        supported_compressions.split(b'|')
      )
    
    elif opcode == toku_c.TOKU_OP_HELLO_ACK:
      payload = _get_payload_from_decode_buffer(&self.decode_buffer)
      selected_encoding, selected_compressor = payload.split(b'|')

      response = opcodes.HelloAck(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_ping_interval(&self.decode_buffer),
        selected_encoding,
        selected_compressor
      )

    elif opcode == toku_c.TOKU_OP_ERROR:
      response = opcodes.Error(
        toku_c.toku_get_flags(&self.decode_buffer),
        toku_c.toku_get_code(&self.decode_buffer),
        toku_c.toku_get_seq(&self.decode_buffer),
        _get_payload_from_decode_buffer(&self.decode_buffer)
      )
    
    self._reset_decode_buf()
    return response

  cdef inline _reset_decode_buf(self):
    toku_c.toku_decoder_reset(&self.decode_buffer)
    _free_big_buffer(&self.decode_buffer.toku_buffer)
    _reset_buffer(&self.decode_buffer.toku_buffer)

  cdef _reset_or_compact_write_buf(self):
    if self.write_buffer_position == self.write_buffer.length:
      _free_big_buffer(&self.write_buffer)
      _reset_buffer(&self.write_buffer)
      self.write_buffer_position = 0
    
    elif self.write_buffer.length > self.write_buffer_position > self.write_buffer.allocated_size / 2:
      memcpy(
        self.write_buffer.buf,
        self.write_buffer.buf + self.write_buffer_position,
        self.write_buffer.length - self.write_buffer_position
      )

      self.write_buffer.length -= self.write_buffer_position
      self.write_buffer_position = 0




cdef class SocketWatcher:
    cdef bint sock_read_ready
    cdef bint sock_write_ready
    cdef bint sock_write_blocked
    cdef bint is_switching
    cdef int sock_fileno
    cdef object hub
    cdef object waiter
    cdef object waiter_switch
    cdef object waiter_get
    cdef object waiter_clear

    cpdef void mark_ready(self, int fileno, bint read=?)
    cpdef request_switch(self)
    cpdef _request_switch(self)

    cdef reset(self)
    cdef wait(self)
    cdef switch_if_write_unblocked(self)



class TokuException(Exception):
  pass

class TokuDecoderError(TokuException):
  pass

class NoEncoderAvailable(TokuException):
  pass

class ConnectionError(TokuException):
  pass

class StreamDefunct(TokuException):
  pass

class ConnectionTerminated(ConnectionError):
  pass

class ConnectionPingTimeout(ConnectionError):
  pass

class InvalidSendException(TokuException):
  pass

class NotClientException(InvalidSendException):
  pass

class NotServerException(InvalidSendException):
  pass

class TokuErrorReceived(TokuException):
  pass




from libc.stdint cimport uint32_t, uint8_t, uint16_t

cdef extern from "../../c/buffer.h":
    ctypedef struct toku_buffer_t:
        char *buf
        size_t length
        size_t allocated_size

    ctypedef struct toku_decode_buffer_t:
        toku_buffer_t toku_buffer;
        uint8_t opcode;
        size_t data_size_remaining;
        size_t header_size;

cdef extern from "../../c/encoder.h":
    int toku_append_hello(toku_buffer_t *b, uint8_t flags, uint32_t size, const char *data)
    int toku_append_hello_ack(toku_buffer_t *b, uint8_t flags, uint32_t ping_interval, uint32_t size, const char *data)
    int toku_append_ping(toku_buffer_t *b, uint8_t flags, uint32_t seq)
    int toku_append_pong(toku_buffer_t *b, uint8_t flags, uint32_t seq)
    int toku_append_request(toku_buffer_t *b, uint8_t flags, uint32_t seq, uint32_t size, const char *data)
    int toku_append_response(toku_buffer_t *b, uint8_t flags, uint32_t seq, uint32_t size, const char *data)
    int toku_append_push(toku_buffer_t *b, uint8_t flags, uint32_t size, const char *data)
    int toku_append_goaway(toku_buffer_t *b, uint8_t flags, uint16_t code, uint32_t size, const char *data)
    int toku_append_error(toku_buffer_t *b, uint8_t flags, uint16_t code, uint32_t seq, uint32_t size, const char *data)

cdef extern from "../../c/decoder.h":
    toku_decoder_status toku_decoder_read_data(toku_decode_buffer_t *pk, size_t size, const char *data, size_t* consumed)
    toku_decoder_status toku_decoder_reset(toku_decode_buffer_t *pk)
    uint32_t toku_get_seq(toku_decode_buffer_t *pk)
    size_t toku_get_data_payload_size(toku_decode_buffer_t *pk)
    uint8_t toku_get_version(toku_decode_buffer_t *pk)
    uint8_t toku_get_flags(toku_decode_buffer_t *pk)
    uint16_t toku_get_code(toku_decode_buffer_t *pk)
    uint32_t toku_get_ping_interval(toku_decode_buffer_t *pk)

cdef extern from "../../c/constants.h":
    const unsigned char TOKU_VERSION;
    const size_t TOKU_DATA_SIZE_MAX;

    ctypedef enum toku_opcodes:
        TOKU_OP_HELLO
        TOKU_OP_HELLO_ACK
        TOKU_OP_PING
        TOKU_OP_PONG
        TOKU_OP_REQUEST
        TOKU_OP_RESPONSE
        TOKU_OP_PUSH
        TOKU_OP_GOAWAY
        TOKU_OP_ERROR

    ctypedef enum toku_decoder_status:
        TOKU_DECODE_NEEDS_MORE
        TOKU_DECODE_COMPLETE
        TOKU_DECODE_MEMORY_ERROR
        TOKU_DECODE_INVALID_OPCODE
        TOKU_DECODE_INVALID_SIZE

    ctypedef enum toku_flags:
        TOKU_FLAG_COMPRESSED



import inspect
import gevent
from gevent.hub import Waiter
from greenlet import getcurrent

cdef class SocketWatcher:
    def __cinit__(self, int sock_fileno):
        self.sock_fileno = sock_fileno
        self.sock_write_blocked = False
        self.waiter = Waiter()
        self.waiter_get = self.waiter.get
        self.waiter_clear = self.waiter.clear
        self.waiter_switch = lambda: self.waiter.switch(None)
        self.hub = gevent.get_hub()
        self.is_switching = False
        self.reset()

    cpdef void mark_ready(self, int fileno, bint read=True):
        if fileno != self.sock_fileno:
            return

        if read:
            self.sock_read_ready = True

        else:
            self.sock_write_ready = True

        self.waiter_switch()

    cdef reset(self):
        self.sock_read_ready = False
        self.sock_write_ready = False
        self.is_switching = False
        self.waiter_clear()

    cdef wait(self):
        return self.waiter_get()

    cpdef request_switch(self):
        if self.is_switching:
            return

        if self.hub is getcurrent():
            self._request_switch()
        else:
            self.hub.loop.run_callback(self._request_switch)

    cpdef _request_switch(self):
        if self.is_switching:
            return

        self.is_switching = True
        self.waiter_switch()

    cdef switch_if_write_unblocked(self):
        if not self.sock_write_blocked:
            self.request_switch()



from __future__ import absolute_import
from __future__ import print_function
import socket
from gevent.server import StreamServer

from toku.encoders import ENCODERS
from toku.socket_session import TokuSocketSession


class TokuServer:
    def __init__(self, server_address, encoders=ENCODERS):
        self._encoders = encoders
        self.server = StreamServer(server_address, self._handle_connection)

    def serve_forever(self):
        self.server.serve_forever()

    def start(self):
        self.server.start()

    def stop(self):
        self.server.stop()

    def _handle_connection(self, sock, addr):
        print('handling connection from', sock)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        sock.setblocking(False)
        session = TokuSocketSession(sock, self._encoders, False, self.handle_request, self.handle_push)

        self.handle_new_session(session)
        try:
            session.join()
        finally:
            self.handle_session_gone(session)

    def handle_request(self, request, session):
        pass

    def handle_push(self, push, session):
        pass

    def handle_new_session(self, session):
        pass

    def handle_session_gone(self, session):
        pass



import socket

import gevent
import time
import six.moves.queue
from gevent.lock import RLock
from gevent.event import AsyncResult

from exceptions import ConnectionError
from socket_session cimport TokuSocketSession
from exponential_backoff cimport Backoff
import logging

from encoders import ENCODERS

cdef class TokuClient:
    cdef TokuSocketSession _session
    cdef object _session_lock
    cdef object _push_handler
    cdef object _connect_greenlet
    cdef object _push_queue
    cdef tuple _address
    cdef object _encoders
    cdef int _connect_timeout
    cdef Backoff _backoff

    def __init__(self, address, push_handler=None, connect_timeout=5, max_push_queue_len=5000, encoders=ENCODERS):
        self._session = None
        self._encoders = encoders
        self._session_lock = RLock()
        self._address = address
        self._connect_timeout = connect_timeout
        self._push_handler = push_handler
        self._connect_greenlet = None
        self._push_queue = six.moves.queue.deque(maxlen=max_push_queue_len)
        self._backoff = Backoff(min_delay=0.25, max_delay=2)

    cpdef set_push_handler(self, object push_handler):
        self._push_handler = push_handler
        if self._session:
            self._session.set_push_handler(push_handler)

    cdef inline _clear_current_session(self):
        cdef TokuSocketSession session = self._session
        with self._session_lock:
            if session is self._session:
                self._backoff.fail()
                self._session = None

    cdef inline _connect_async(self):
        # If we are connected and ready, return true so that we send to the session right away.
        if self._session is not None and not self._session.defunct() and self._session.is_ready():
            return True

        # Otherwise, if the session is established, but not yet ready, return false so that we append
        # to the queue, which will be consumed when we're ready.
        if self._session is not None and not self._session.defunct() and not self._session.is_ready():
            return False

        # Otherwise, we need to connect. Spawn a greenlet to do the connection.
        if self._connect_greenlet is None:
            self._connect_greenlet = gevent.spawn(self._connect_in_greenlet)

        # Return false to signify we should append to the queue.
        return False

    cpdef _connect_and_send_request_async(self, request_data):
        self.connect()
        response = self._session.send_request(request_data)
        return response.get()

    cdef inline _is_socket_connected(self):
        return self._session is not None and not self._session.defunct()

    def _connect_in_greenlet(self):
        try:
            self.connect()
        finally:
            self._connect_greenlet = None

    cdef connect(self):
        cdef TokuSocketSession session

        if self._session is not None:
            if self._session.defunct():
                self._clear_current_session()
            else:
                return

        with self._session_lock:
            if self._session is not None:
                return

            if self._backoff.fails():
                gevent.sleep(self._backoff.current())

            start = time.time()

            try:
                logging.info('[toku %s:%s] connecting', self._address[0], self._address[1])
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                sock.settimeout(self._connect_timeout)
                sock.connect(self._address)
                self.handle_new_socket(sock)
                sock.setblocking(False)
                session = TokuSocketSession(sock, encoders=self._encoders, on_push=self._push_handler)
                elapsed = (time.time() - start) * 1000
                logging.info('[toku %s:%s] connected in %.2f ms.', self._address[0], self._address[1], elapsed)
                session.await_ready()

                if not session.defunct() and self._push_queue:
                    logging.info('[toku %s:%s] flushing %i push events.', self._address[0], self._address[1],
                                 len(self._push_queue))

                    while self._push_queue:
                        session.send_push(self._push_queue.popleft())

                self._backoff.succeed()
                self._session = session
            except socket.error:
                logging.exception('[toku %s:%s] connection failed after %.2f ms.',
                                  self._address[0], self._address[1], (time.time() - start) * 1000)
                self._backoff.fail()
                raise ConnectionError('No connection available')

    cpdef send_request_async(self, request_data):
        if not self._is_socket_connected():
            async_result = AsyncResult()
            gevent.spawn(self._connect_and_send_request_async, request_data, async_result).rawlink(async_result)
            return async_result

        return self._session.send_request(request_data)

    cpdef send_request(self, request_data, timeout=None, retries=3):
        try:
            self.connect()
            response = self._session.send_request(request_data)
            return response.get(timeout=timeout)
        except ConnectionError:
            if retries:
                return self.send_request(request_data, timeout, retries - 1)

            raise

    cpdef send_push(self, push_data):
        if not self._connect_async():
            self._push_queue.append(push_data)

        else:
            self._session.send_push(push_data)

    def handle_new_socket(self, socket):
        pass


cdef class TokuHTTPUpgradeClient(TokuClient):
    def handle_new_socket(self, sock):
        upgrade_payload = b'\r\n'.join([
            b'GET /_rpc HTTP/1.1',
            (b'Host: %s' % six.ensure_binary(self._address[0], 'ascii')),
            b'Upgrade: toku',
            b'Connection: Upgrade',
            b'',
            b''
        ])

        sock.sendall(upgrade_payload)
        expected_handshake_responses = [
            b'\r\n'.join([
                b'HTTP/1.1 101 Switching Protocols',
                b'Upgrade: toku',
                b'Connection: Upgrade',
                b'',
                b''
            ]).lower(),
            b'\r\n'.join([
                b'HTTP/1.1 101 Switching Protocols',
                b'Connection: Upgrade',
                b'Upgrade: toku',
                b'',
                b''
            ]).lower(),
        ]

        bytes_remaining = len(expected_handshake_responses[0])
        handshake_response = b''
        while bytes_remaining:
            buf = sock.recv(bytes_remaining)
            if not buf:
                raise ConnectionError('Connection died while reading handshake response')

            handshake_response += buf
            bytes_remaining -= len(buf)

        if handshake_response.lower() not in expected_handshake_responses:
            raise ConnectionError('Invalid handshake response: %s' % handshake_response)



from libc.stdint cimport uint32_t, uint8_t, uint16_t
cimport toku_c

cdef class TokuStreamHandler:
  cdef uint32_t seq
  cdef toku_c.toku_decode_buffer_t decode_buffer
  cdef toku_c.toku_buffer_t write_buffer
  cdef size_t write_buffer_position

  cdef inline uint32_t next_seq(self)
  cpdef uint32_t current_seq(self)

  cpdef uint32_t send_ping(self, uint8_t flags) except 0
  cpdef uint32_t send_pong(self, uint8_t flags, uint32_t seq) except 0
  cpdef uint32_t send_request(self, uint8_t flags, bytes data) except 0
  cpdef uint32_t send_push(self, uint8_t flags, bytes data) except 0
  cpdef uint32_t send_response(self, uint8_t flags, uint32_t seq, bytes data) except 0
  cpdef uint32_t send_hello(self, uint8_t flags, list supported_encodings, list supported_compressors) except 0
  cpdef uint32_t send_hello_ack(self, uint8_t flags, uint32_t ping_interval, bytes selected_encoding, bytes selected_compressor) except 0
  cpdef uint32_t send_error(self, uint8_t flags, uint16_t code, uint32_t seq, bytes data) except 0
  cpdef uint32_t send_goaway(self, uint8_t flags, uint16_t code, bytes reason) except 0

  cpdef size_t write_buffer_len(self)
  cpdef bytes write_buffer_get_bytes(self, size_t length, bint consume=?)
  cpdef size_t write_buffer_consume_bytes(self, size_t length)
  cpdef list on_bytes_received(self, bytes data)

  cdef object _consume_decode_buffer(self)
  cdef inline _reset_decode_buf(self)
  cdef _reset_or_compact_write_buf(self)



from libc.stdint cimport uint32_t, uint8_t

from stream_handler cimport TokuStreamHandler
from socket_watcher cimport SocketWatcher
from opcodes cimport Ping, Pong, Hello, Request, Response, Push, HelloAck, GoAway, Error

cdef class TokuSocketSession:
    cdef TokuStreamHandler _stream_handler
    cdef object _sock
    cdef SocketWatcher _watcher
    cdef dict _inflight_requests
    cdef bint _is_client

    cdef bint _shutting_down
    cdef object _shutdown_event

    cdef object _close_event

    cdef bint _is_ready
    cdef object _ready_event

    cdef uint32_t _ping_interval
    cdef object _available_encoders
    cdef object _available_compressors

    cdef object _on_request
    cdef object _on_push
    cdef object _encoder_loads
    cdef object _encoder_dumps
    cdef bytes _encoding

    cpdef set_push_handler(self, object push_handler)
    cpdef join(self, timeout=?)
    cpdef terminate(self)
    cpdef close(self, uint8_t code=?, bytes reason=?, block=?, block_timeout=?, close_timeout=?,
                bint via_remote_goaway=?)

    cdef bint defunct(self)
    cdef bint await_ready(self) except 1
    cdef bint is_ready(self)

    cdef _resume_sending(self)
    cpdef _close_timeout(self)
    cdef shutdown(self)
    cdef void _cleanup_socket(self)
    cdef _cleanup_inflight_requests(self, close_exception)
    cdef _encode_data(self, object data)
    cdef _decode_data(self, uint8_t flags, object data)
    cpdef object send_request(self, object data)
    cpdef object send_push(self, object data)
    cpdef object send_response(self, uint32_t seq, object data)
    cpdef object send_ping(self)
    cpdef object _send_hello(self)
    cpdef object _send_hello_ack(self, bytes selected_encoding, bytes selected_compressor)

    cdef _handle_data_received(self, data)
    cdef _handle_ping_timeout(self)

    cdef _handle_client_events(self, list events)
    cdef _handle_server_events(self, list events)

    cdef _handle_push(self, Push push)
    cdef _handle_request(self, Request request)
    cdef _handle_response(self, Response response)

    cdef _handle_ping(self, Ping ping)
    cdef _handle_pong(self, Pong pong)

    cdef _handle_go_away(self, GoAway go_away)
    cdef _handle_error(self, Error error)

    cdef _handle_hello(self, Hello hello)
    cdef _handle_hello_ack(self, HelloAck select_encoding)

    cdef _pick_best_encoding(self, list encodings)
    cpdef _ping_loop(self)
    cpdef _run_loop(self)



cdef class Backoff:
  cdef float _min
  cdef float _max
  cdef float _current
  cdef int _fails
  cdef bint _jitter

  cdef int fails(self)
  cdef float current(self)
  cdef void succeed(self)
  cdef float fail(self)



import os
from setuptools import setup
from setuptools.extension import Extension
import glob

try:
  from Cython.Build import cythonize
  ext = "pyx"
except ImportError:
  cythonize = None
  ext = "c"

exts = []
for file in glob.glob("py/toku/*.%s" % ext):
  package = os.path.splitext(os.path.basename(file))[0]
  exts.append(Extension(
    "toku.%s" % package,
    [file],
    extra_compile_args=["-O3"]
  ))

if cythonize:
  exts = cythonize(exts)

setup(
  name='toku',
  version='0.2.23',
  author='Constanze',
  author_email='cstanze@helselia.dev',
  url='https://github.com/Helselia/Toku',
  description='A simple stream based RPC - with a gevent client/server implementation. A fork of discord/loqui',
  license='MIT',
  package_dir={
    '': 'py'
  },
  packages=['toku'],
  ext_modules=exts
)
